/*
 * Copyright (C) 1997.  Netscape Communications Corp. 
 *
 * Unsupported Example Code.
 * 
 * All examination and use of this code is covered by the
 * unsupported software license agreement accompanying it.
 *
 */


/*
 * This is example code for a simple VGI program that
 * will authenticate users based on an existing NT password source
 * scheme on the certificate server host.  It is an example for
 * illustrative purposes only.   This should not be considered deployment
 * quality code nor an indication of any recommendation for a deployment.
 * 
 * Note in particular that it is not recommended to use insecure network-based
 * password sources on a secure certificate server host, and
 * that in an actual deployment it is advisable to use a different
 * source of password information than that used for login on the
 * certificate server host machine.
 *
 */

/*
 * Using the Microsoft Visual C++ 2.0 compiler this will
 * compile with the command:
 * % cl -MD -Fentexample.exe ntexample.c -I../include ../lib/libvgiutil.lib netapi32.lib advapi32.lib
 *
 * If compiling from within the Developer Studio environment, add
 * these libraries to the settings for link options.  It has been
 * compiled and tested using MSVC++ 2.2, and MSVC++ 4.0 compilers.
 */


#include <windows.h>
#include <lm.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <io.h>

	
/*
 * This is missing from winbase.h on many older developer systems.
 * You can find this by searching for LOGON_NETWORK at
 * www.microsoft.com under MSDN
 */

#ifndef LOGON32_LOGON_NETWORK
#define LOGON32_LOGON_NETWORK 3
#endif     

#include "pblock.h"
#include "urlcod.h"

LPUSER_INFO_2
getUserInfo(char *userName, char *domain, NET_API_STATUS *nstatus)
{
    LPWSTR dcName=0;
    LPUSER_INFO_2 uip=0;
    
    if (domain) {
	WCHAR wszDomain[256]; 
	MultiByteToWideChar(CP_ACP, 0, domain, strlen(domain)+1,
			    wszDomain,
			    sizeof(wszDomain)/sizeof(wszDomain[0]) );

	/*
	 * Get the name of a domain controller for the domain.  If this
	 * fails, it leaves dcName == 0, which indicates the localhost as
	 * the server to process the next call.
	 */
	NetGetDCName( NULL, wszDomain, (LPBYTE *) &dcName );
    }


    {
	/* Look up the user on the domain controller. */
	/* Convert ASCII user name and domain to Unicode */
	WCHAR wszUserName[256];
	MultiByteToWideChar(CP_ACP, 0, userName, strlen(userName)+1,
			    wszUserName,
			    sizeof(wszUserName)/sizeof(wszUserName[0]) );

	*nstatus = NetUserGetInfo(dcName,  (LPWSTR) wszUserName, 2,
				  (LPBYTE *) &uip);
    }

    /* Free dcName */
    if (dcName) NetApiBufferFree(dcName);

    return (uip);
}
 


static
void
setBinaryMode(FILE *fp)
{
    /* This makes content length calculations easier by telling Windows
     * not to mess with LF->CR/LF conversions
     */
    _setmode(_fileno(fp),_O_BINARY);
}


char *progName = 0;


/*
 * This routine is used to send a crude error message back to the client.
 *
 * In actual practice you can send arbitrary HTML and better error messages.
 *
 */

void
gatewayError(const char *msg)
{
    printf("Content-Type: text/html\n");
    printf("Content-Length: %d\n", strlen(progName) + 2 + strlen(msg));
    printf("\n");
    printf("%s: %s", progName, msg);
    exit(0);
}    


/* Arbitrary -- you can use allocation on heap (malloc()) to avoid this */ 
#define INBUFSIZE 8192

/* Varies -- may be in limits.h.  Maximum allowed length of a user name */
#ifndef LOGNAME_MAX
#define LOGNAME_MAX 32
#endif

/* Maximum length of a subject name */
#define SUBJECT_MAX 255
#define REALNAME_MAX 64

/* Validity period for certs issued - 6 months (180 days) */
#define DAY 86400
#define VALIDITY_PERIOD (6*30*DAY)


int
main(int argc, char **argv, char **envp)
{
    char *contentLenStr = getenv("CONTENT_LENGTH");
    char inBuf[INBUFSIZE];
    int contentLen;
    int actualLen;
    pblock *inArgs;
    pblock *outArgs;
    char *usernameRaw;
    char username[LOGNAME_MAX+1];
    char realname[REALNAME_MAX+1];
    char *domain=0;
    char validityBuf[64];
    char *password;
    char subject[SUBJECT_MAX];
    char *spki;
    char *outForm;
    LPUSER_INFO_2 userInfo=0;
    NET_API_STATUS nstatus=0;
    BOOL userAuthenticated=FALSE;
    HANDLE userHandle=0;
    int logonType=0;

    progName = argv[0];

    if (contentLenStr == 0) {
	printf("This program is a VGI program; it expects to be\n"
	       "run through the certificate server auto-verification\n"
	       "interface\n");
	exit(1);
    }
    
    contentLen = atoi(contentLenStr);
    if (contentLen <= 0) {
	gatewayError("Ill-formed contentlength from gateway.");
    }
    if (contentLen > INBUFSIZE-1) {
	gatewayError("Size of submission is too large");
    }
    
    actualLen = 0;
    while (actualLen < contentLen) {
	int c = getc(stdin);
	if (c == EOF) break;
	inBuf[actualLen] = (char) c;
	actualLen++;
    }
    inBuf[actualLen] = 0;    

    if (actualLen < contentLen) {
	gatewayError("Actual length less than stated content length.");
    }
    
    inArgs = url_encoding_to_pblock(inBuf);
    if (inArgs == 0) {
	gatewayError("Input to VGI program was improperly formed.");
    }

    /*
     * Map input to a subject name and authenticate as you wish.
     * Here the form submission must include a "username" and "password"
     * and we use the machine's source of user password information
     * to authenticate and formulate the subject name.
     */

    usernameRaw = pblock_findval("username", inArgs);
    if (usernameRaw == 0) {
	gatewayError("Rejected. No username attribute in request.");
    }
    if (strlen(usernameRaw) < 1 || strlen(usernameRaw) > LOGNAME_MAX) {
	gatewayError("Rejected.   Username is empty or too long.");
    }
    {
	/* convert user name to lower case */
	char *s, *t;
	for (s = usernameRaw, t=&username[0]; *s != 0; s++, t++ ) {
	    *t = (char) tolower(*s);
	}
	*t = 0;
    }
    

    password = pblock_findval("passwd", inArgs);
    if (password == 0) {
	gatewayError("Rejected. No password attribute in request.");
    }

    spki = pblock_findval("subjectPublicKeyInfo",inArgs);
    if (spki == 0) {
	gatewayError("Rejected.  No public key info in request.");
    }

    /* 
     * This environment variable can be supplied on the obj.conf
     * autoVerifier config line by specifying lmdomain = "<lanmanagerdomain>". 
     */
    domain = getenv("LMDOMAIN");
	
    /*
     * Use LogonUser to check the username/password supplied
     * This is complicated by a hack to get around the fact that
     * some systems don't support the more efficient LOGON_NETWORK logon
     * mode and depending on how the VGI is run, it may not be able to
     * do the LOGON_BATCH.   If you know one that you can use
     * on your system, you can eliminate the for-loop and save
     * runtime costs.
     */
    
    {
	int i;
	int logonTypesToTry[] = {
	    LOGON32_LOGON_NETWORK,
	    LOGON32_LOGON_BATCH,
	    LOGON32_LOGON_INTERACTIVE,
	    0
        };
	for (i=0; logonTypesToTry[i] != 0; i++) { 
	    userAuthenticated = LogonUser(username, domain, password,
					  logonTypesToTry[i],
					  LOGON32_PROVIDER_DEFAULT,
					  &userHandle);
	    if (!userAuthenticated) {
		int err = GetLastError();
		switch (err) {
		  case ERROR_LOGON_FAILURE:
		    gatewayError("Invalid username or password. "
				 "Check your typing and try again.");
		    /* does not return */
		    break;
		  case ERROR_LOGON_TYPE_NOT_GRANTED:
		  case ERROR_INVALID_PARAMETER:
		  default:
		    /* The logon type requires NT privileges this process
		     * doesn't have or the logon type is not available on
		     * this system.  Continue looping trying next logon type.
		     */
		    break;
		}
	    }
	    if (userHandle) {
		CloseHandle(userHandle);
		userHandle = 0;
	    }
	}
	if (!userAuthenticated) {
	    /* All logon types resulted in system-type failures */
	    gatewayError("LogonUser() system-failures for all listed logon "
			 "types.");
	    /* does not return */
	}
    }


    /* Get the user info */
    userInfo = getUserInfo(username, domain, &nstatus);
    if (userInfo ==0) {
	gatewayError("Invalid username or password. "
		     "Check your typing and try again.");
	/* does not return */
    }


    /* Get the real name from the USER_INFO */
    if (!WideCharToMultiByte(CP_ACP, 0,
			     userInfo->usri2_full_name, -1,
			     (LPSTR) realname, REALNAME_MAX,
			     0, 0)) {
	gatewayError("Unable to convert user's full name to ASCII.");
	/* does not return */
    }

    /* Free the userInfo */
    NetApiBufferFree(userInfo);


    /* OK.  We have authenticated the user and have the information we want
     * for formulating the certificate, which we proceed to do.
     */

    /* Allocate pblock for holding parameters of issue operation */
    outArgs = pblock_create(16);
    if (outArgs == 0) {
	gatewayError("Unable to create pblock.");
	/* does not return */
    }

    /*
     * Form subject name from user's login, real name, and a
     * fixed organization name.  This should be done in a way that
     * matches your directory structure.  The name formulation here 
     * may or may not match your organization.  You should change
     * it so that it does.  You can do fancier things like determine
     * an organization unit (OU) from the user's primary group id or
     * you can use the LDAP SDK itself and use an existing password
     * in the directory for the initial authentication.
     */
    {
	/*
	 * Note -- this will fit in subject[] without danger of overflow.
	 * If you change this you may have to check it.
	 */
	char *organization = "Test Corporation";
	char *country = "US";
	char *subjectFormat = "UID=%s, CN=%s, O=%s, C=%-2s";
	sprintf(subject, subjectFormat,
		username,  realname, organization, country);
    }

    /* See the documentation for more information on how to formulate
     * an issue request and how to specify additional extensions.
     */

    /* Generate string decimal form for validity period */ 
    sprintf(validityBuf, "%d", VALIDITY_PERIOD);

    /* Generate parameters for issue request */
    if (!pblock_nvinsert("op", "issue", outArgs) ||
	!pblock_nvinsert("subjectPublicKeyInfo", spki, outArgs) ||
	!pblock_nvinsert("subject", subject, outArgs) ||

	/* validity */
	!pblock_nvinsert("validityLength", validityBuf, outArgs) ||

	/* instruct server to add these extensions if extensions are enabled */
	!pblock_nvinsert("AuthorityKeyIdentifier", "", outArgs) ||
	!pblock_nvinsert("certTypeSSLClient", "true", outArgs) ||
	!pblock_nvinsert("certTypeEmail", "true", outArgs)) {

	/* failed unexpectedly */
	gatewayError("Failure while setting up issue parameters.");
	/* does not return */
    }
	
    /* URL-encode the parameters */
    outForm = pblock_to_url_encoding(outArgs);
    if (outForm == 0) {
	gatewayError("Failed url-encoding issue request.");
	/* does not return */
    }

    /* Output it back to the server */
    printf("Content-Type: application/x-www-form-urlencoded\n");
    printf("Content-Length: %d\n", strlen(outForm));
    printf("\n");
    printf("%s", outForm);

    return(0);
}
