/*
 * Copyright (C) 1997.  Netscape Communications Corp. 
 *
 * Unsupported Example Code.
 * 
 * All examination and use of this code is covered by the
 * unsupported software license agreement accompanying it.
 *
 */

/*
 * urlcod.c: translation of pblock to/from URL-encoded form.
 * See urlcod.h for more information
 *
 */


#include "pblock.h"
#include "urlcod.h"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>


static int
hexToInt(int digit)
{
    int value = 0;
    
    if(digit >= '0' && digit <= '9') {
	value = digit - '0';
    } else if(digit >= 'A' && digit <= 'F' ) {
	value = digit - 'A' + 10;
    } else if(digit >= 'a' && digit <= 'f' ) {
	value = digit - 'a' + 10;
    }
    return value;
}


static int
intToHex(int number)
{
    int digit = '0';
    
    if(number >= 0 && number <= 9) {
	digit = number + '0';
    } else if(number >= 10 && number <= 15 ) {
	digit = number - 10 + 'A';
    }
    return digit;
}


static
int
removeSeparators(char *src, char *dst, char separator)
{
    if ( src == 0 || dst == 0 ) {
	return 0;
    }
    for( ; *src; ++src, ++dst ) {
	if(*src == separator) {
	    *dst = ' ';
	} else {
	    *dst = *src;
	}
    }
    *dst = '\0';
    return 1;
}


static
int
parseString(char **str, char **n, char **v)
{
    char *string = *str;
    char *name = *n;
    char *value = *v;
    
    if (*string == '\0') {
	/* string is empty */
	name = 0; value = 0;
	*n = name;
	*v = value;
	return 1;
    }
    
    /* find end of name */
    name = string;
    while(*string && *string != ' ' && *string != '=') {
	string++;
    }
    
    /* name found, look for value */
    if (*string == '=') {
	/* terminate name */
	*string++ = '\0';
	value = string;
	while(*string && *string != ' ') {
	    string++;
	}
    } else {
	value = "";
    }
    
    /* zero-terminate/increment for next name/value */
    if( *string ) {
	*string++ = '\0';
    }
    *str = string;
    *n = name;
    *v = value;
    return 1;
}



static
int
removeEscapes(const char *src, char *dst)
{
    if ( src == 0 || dst == 0 ) {
	return 0;
    }
    for( ; *src ; ++src, ++dst ) {
	if ( *src != '%' ) {
	    if (*src == '+') {
		*dst = ' ';
	    } else {
		*dst = *src;
	    }
	} else {
	    if ( ! isxdigit(src[1]) || ! isxdigit(src[2]) ) {
		/* ill-formed */
		return 0;
	    }
	    /* read hex number into value */
	    *dst = hexToInt(src[1]) * 16 + hexToInt(src[2]);
	    src += 2;
	}
    }
    *dst = '\0';
    return 1;
}


static
int
insertEscapes(const char *src, char *dst)
{
    char *d = dst;
    for( ; *src; ++src) {
	if(*src == ' ') {
	    *d++ = '+';
	}
	else if(strchr((const char*)"%+=\n\r\t;&", (int)(*src))) {
	    *d++ = '%';
	    *d++ = intToHex((*src)>>4);
	    *d++ = intToHex((*src)&0x0F);
	} else {
	    *d++ = *src;
	}
    }
    *d = '\0';
    return (int) (d - dst);
}







/*  Parses url-encoded string into pblock structure */

pblock *
url_encoding_to_pblock(const char *s)
{
    pblock *pb;
    char *buffer, *b;

    if ( s == 0 ) {
	return 0;
    }
    
    pb = pblock_create(16); 
    if (pb == 0) {
	return 0;
    }

    if (*s == 0) {
	return pb;
    }

    buffer = strdup(s);
    if (buffer == 0) {
	pblock_free(pb);
	return 0;
    }
    
    removeSeparators(buffer, buffer, '&');
    b = buffer;
    do {
	char *n=0, *v=0;
	if (!parseString(&b, &n, &v)) {
	    free(buffer);
	    pblock_free(pb);
	    return 0;
	}

	if (n == 0 || v == 0) {
	    break;
	}

	if (((*v != 0) && !removeEscapes(v, v)) ||
	    (pblock_nvinsert(n,v,pb) == 0)) {
	    free(buffer);
	    pblock_free(pb);
	    return 0;
	}
    } while ( *b );

    free(buffer);
    return pb;
}




char *
pblock_to_url_encoding(pblock *pb)
{
    pb_iter i;
    pb_param *pp;
    char *buffer, *b;
    int len;

    if (pb == 0) {
	return 0;
    }

    /* calculate maximum possible length -- could be more efficient */
    len = 0;
    for (pb_iter_init(&i, pb); !pb_iter_at_end(&i); pb_iter_next(&i)) {
	pp = pb_iter_cur(&i);
	len += strlen(pp->name) + 1 + 3*strlen(pp->value) + 1;
	/* 1 for '=', 3 for escape expansion, 1 for '&' or ending null */
    }

    buffer = (char *) malloc(len);
    if (buffer == 0) {
	return 0;
    }

    b = buffer;
    pb_iter_init(&i, pb);
    while (!pb_iter_at_end(&i)) {
	char *n;
	pp = pb_iter_cur(&i);
	n = pp->name;
	while (*n) *b++ = *n++;
	*b++ = '=';
	b += insertEscapes(pp->value,b);
	pb_iter_next(&i);
	if (!pb_iter_at_end(&i)) {
	    *b++ = '&';
	}
    }

    return buffer;
}		


